;;;; macosx-env.scm
;;;; Kon Lovett, Sep '06

;; Issues
;;
;; - Assumes UTF8 encoding for SCDynamicStoreCopyComputerName
;;
;; - No SCDynamicStoreCopyProxies support
;;
;; - No CFStringGetCStringPtr use since memory allocation is not performed.
;; This is sub-par but makes the Scheme invocation more complex.

(declare
  (no-procedure-checks)
	(no-bound-checks)
	(export
	  macosx:session-info
		macosx:machine-name
		macosx:short-user-name
		macosx:long-user-name
		macosx:machine-location
		macosx:metric?
		macosx:computer-name
		macosx:console-user
		macosx:local-host-name
		macosx:location-name
		macosx:main-bundle-path
		macosx:tick-count
		macosx:delay
		macosx:gestalt
		;
    macosx:get-bundle-path) )

;;;

#>
#include <math.h>
#include <unistd.h>

#include <CoreFoundation/CoreFoundation.h>
#include <CoreServices/CoreServices.h>
#include <SystemConfiguration/SCDynamicStoreCopySpecific.h>
#include <ApplicationServices/ApplicationServices.h>
#include <Security/AuthSession.h>

#ifdef HAVE_STDINT_H
# include <stdint.h>
#else
# include <sys/types.h>
  typedef u_int8_t  uint8_t;
  typedef u_int16_t uint16_t;
  typedef u_int32_t uint32_t;
  typedef u_int64_t uint64_t;
#endif

#define VECLEN(v) (sizeof(v)/sizeof((v)[0]))

static void
cfsr_to_utf8str (CFStringRef cfsr, unsigned char **outstr)
{
#	define LONG_CHARS_PER_UTF8 6
	CFIndex buflen = LONG_CHARS_PER_UTF8 * CFStringGetLength (cfsr) + 1;
	*outstr = ((unsigned char *) C_malloc (buflen));
	if (! CFStringGetCString (cfsr, (char *) *outstr, buflen, kCFStringEncodingUTF8)) {
		C_free (*outstr);
		*outstr = NULL;
	}
#	undef LONG_CHARS_PER_UTF8
}

static void
machine_name (unsigned char **outstr)
{
	cfsr_to_utf8str (CSCopyMachineName(), outstr);
}

static void
short_user_name (unsigned char **outstr)
{
	cfsr_to_utf8str (CSCopyUserName (true), outstr);
}

static void
long_user_name (unsigned char **outstr)
{
	cfsr_to_utf8str (CSCopyUserName (false), outstr);
}

static void
machine_location (double *lat, double *lon, int *dls, long *gmt)
{
#	define ROUNDN(v,p) (round((v) * ((p) * 10.0)) / ((p) * 10.0))
	MachineLocation machloc;
	ReadLocation (&machloc);
	*lat = ROUNDN (((double) FractToFloat (machloc.latitude)) * 90.0, 4);
	*lon = ROUNDN (((double) FractToFloat (machloc.longitude)) * 90.0, 4);
	*dls = 0 < machloc.u.dls.Delta ? 3600 : (0 > machloc.u.dls.Delta ? -3600 : 0);
	*gmt = ((machloc.u.gmtDelta & 0x00FFFFFF) << 8) >> 8;
#	undef ROUNDN
}

static void
computer_name (unsigned char **outstr, SCDynamicStoreRef store)
{
  /* Assumes UTF8 encoding! */
	CFStringRef cfsr = SCDynamicStoreCopyComputerName (store, NULL);
	cfsr_to_utf8str (cfsr, outstr);
	CFRelease (cfsr);
}

static void
console_user (unsigned char **outstr, uint32_t *puid, uint32_t *pgid, SCDynamicStoreRef store)
{
	uid_t uid;
	gid_t gid;
	CFStringRef cfsr = SCDynamicStoreCopyConsoleUser (store, &uid, &gid);
	cfsr_to_utf8str (cfsr, outstr);
	CFRelease (cfsr);
	*puid = uid;
	*pgid = gid;
}

static void
local_host_name (unsigned char **outstr, SCDynamicStoreRef store)
{
	CFStringRef cfsr = SCDynamicStoreCopyLocalHostName (store);
	cfsr_to_utf8str (cfsr, outstr);
	CFRelease (cfsr);
}

static void
location_name (unsigned char **outstr, SCDynamicStoreRef store)
{
	CFStringRef cfsr = SCDynamicStoreCopyLocation (store);
	cfsr_to_utf8str (cfsr, outstr);
	CFRelease (cfsr);
}

static void
main_bundle_path (unsigned char **outstr)
{
  CFBundleRef bundle = CFBundleGetMainBundle();
  *outstr = NULL;
  if (NULL != bundle) {
    CFURLRef url = CFBundleCopyExecutableURL (bundle);
    if (NULL != url) {
      long buflen = pathconf ("/", _PC_PATH_MAX); /* Any pathname will do */
      *outstr = ((unsigned char *) C_malloc (buflen));
      if (NULL != outstr) {
        if (!CFURLGetFileSystemRepresentation (url, true, *outstr, buflen)) {
          C_free (outstr);
          *outstr = NULL;
        }
      }
    }
  }
}

static uint32_t
string_to_OSType (char *str)
{
  union {uint32_t v; uint8_t c[4];} ost;
  int i;
  for (i = 0; i < VECLEN(ost.c) && *str; ++i, ++str)
    ost.c[i] = (uint8_t) *str;
  for (; i < VECLEN(ost.c); ++i)
    ost.c[i] = (uint8_t) ' ';
  return ost.v;
}

static int
this_session_info_1 (uint32_t *psid, uint32_t *psab)
{
  SecuritySessionId mySession;
  SessionAttributeBits sessionInfo;
  OSStatus error = SessionGetInfo (callerSecuritySession, &mySession, &sessionInfo);
  if (errSessionSuccess == error) {
    *psid = mySession;
    *psab = sessionInfo;
    return 0;
  }
  return error;
}

#define session_LoginCompleted  0x0001
#define session_UserIsActive    0x0010

static void
this_session_info_2 (unsigned char **outstr, uint32_t *puid, uint32_t *pcon, uint32_t *pbit)
{
  CFDictionaryRef sessionInfoDict = CGSessionCopyCurrentDictionary();
  if (NULL != sessionInfoDict) {
    CFStringRef shortUserName = CFDictionaryGetValue (sessionInfoDict, kCGSessionUserNameKey);
    CFNumberRef userUID = CFDictionaryGetValue (sessionInfoDict, kCGSessionUserIDKey);
    CFNumberRef consoleSet = CFDictionaryGetValue (sessionInfoDict, kCGSessionConsoleSetKey);
    CFBooleanRef userIsActive = CFDictionaryGetValue (sessionInfoDict, kCGSessionOnConsoleKey);
    CFBooleanRef loginCompleted = CFDictionaryGetValue (sessionInfoDict, kCGSessionLoginDoneKey);
    CFNumberGetValue (userUID, kCFNumberSInt32Type, puid);
    CFNumberGetValue (consoleSet, kCFNumberSInt32Type, pcon);
    *pbit = (CFBooleanGetValue (loginCompleted) << 1) | CFBooleanGetValue (userIsActive);
    cfsr_to_utf8str (shortUserName, outstr);
  } else {
    *outstr = NULL;
  }
}
<#

;;

(define-macro (call/unsigned-string-result C-NAME . EXTRA)
	`(let-location ([outstr unsigned-c-string*])
	   ($ void ,C-NAME #$outstr ,@EXTRA)
		 outstr ) )

(define-macro (foreign-mask-set? ?bits ?c-nam)
  `(not (zero? (bitwise-and ,?bits (foreign-value ,?c-nam unsigned-integer32)))) )

;;

(define (macosx:session-info)
  (let-location ([sid unsigned-integer32] [sab unsigned-integer32])
    (let ([sta ($ int this_session_info_1 #$sid #$sab)])
      (if (fx= 0 sta)
        (let ([vec (vector sid
                           (foreign-mask-set? sab "sessionIsRoot")
                           (foreign-mask-set? sab "sessionHasGraphicAccess")
                           (foreign-mask-set? sab "sessionHasTTY")
                           (foreign-mask-set? sab "sessionIsRemote")
                           (foreign-mask-set? sab "sessionWasInitialized")
                           #f #f #f #f #f)])
          (let-location ([uid unsigned-integer32] [con unsigned-integer32] [bits unsigned-integer32])
	          (let ([nam (call/unsigned-string-result this_session_info_2 #$uid #$con #$bits)])
              (when nam
                (vector-set! vec 6 nam)
                (vector-set! vec 7 uid)
                (vector-set! vec 8 con)
                (vector-set! vec 9 (foreign-mask-set? bits "session_LoginCompleted"))
                (vector-set! vec 10 (foreign-mask-set? bits "session_UserIsActive")) )
              vec ) ) )
        (error 'macosx:session-info "OSStatus" sta) ) ) ) )

;;

(define (macosx:machine-name)
	(call/unsigned-string-result machine_name) )

(define (macosx:short-user-name)
	(call/unsigned-string-result short_user_name) )

(define (macosx:long-user-name)
	(call/unsigned-string-result long_user_name) )

(define (macosx:machine-location)
	(let-location ([lat double] [lon double] [dls int] [gmt long])
		($ void machine_location #$lat #$lon #$dls #$gmt)
		(vector lat lon dls gmt) ) )

(define (macosx:metric?)
	($ bool IsMetric) )

;;

(define (macosx:computer-name #!optional (store #f))
	(call/unsigned-string-result computer_name (c-pointer store)) )

(define (macosx:console-user #!optional (store #f))
	(let-location ([uid unsigned-integer32] [gid unsigned-integer32])
	  (let ([nam (call/unsigned-string-result console_user #$uid #$gid (c-pointer store))])
	    (and nam
		      (vector nam uid gid) ) ) ) )

(define (macosx:local-host-name #!optional (store #f))
	(call/unsigned-string-result local_host_name (c-pointer store)) )

(define (macosx:location-name #!optional (store #f))
	(call/unsigned-string-result location_name (c-pointer store)) )

;;

(define (macosx:main-bundle-path)
  (call/unsigned-string-result main_bundle_path) )
(define macosx:get-bundle-path macosx:main-bundle-path)

;;

(define (macosx:tick-count)
  ($ unsigned-integer32 TickCount) )

(define (macosx:delay ticks)
	(let-location ([fticks unsigned-long])
    ($ void Delay (unsigned-long ticks) #$fticks)
    fticks ) )

;;

(define (macosx:gestalt sel)
  ; Symbolic selector -> numeric
  (when (symbol? sel)
    (set! sel (symbol->string sel)))
  (when (string? sel)
    (set! sel ($ unsigned-integer32 string_to_OSType (c-string sel))))
  ; Must be numeric
  (unless (and (integer? sel) (positive? sel))
    (error 'macosx:gestalt "invalid selector" sel))
  ;
	(let-location ([resp long])
    (let ([err ($ short Gestalt (unsigned-integer32 sel) #$resp)])
      (if (fx= 0 err)
        resp
        (error 'macosx:gestalt "OSErr" err) ) ) ) )
